# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Module to provide different functions relating to time and time string
    conversion and handling. """

import time, photohandler

class NoImageError(Exception):
  pass

def getSystemUTCOffset(logger, date = None):
  """ Returns the difference in seconds from UTC (east is positive), at the
      local time tuple specified, or now if no time parameter is given. """
  
  # First we need to figure out it DST is active. This is done by reading the
  # 9th parameter of the tuple returned by time.localtime() If a time tuple is
  # given and DST is unknown, we first need to convert it to seconds.
  if ((date) and (date[8] < 0)):
    dst_active = time.localtime(time.mktime(date))[8]
  elif (date):
    dst_active = date[8]
  else:
    dst_active = time.localtime()[8]
    
  # Return the difference with regard to DST
  if dst_active:
    return -1 * time.altzone
  else:
    return -1 * time.timezone
    
def getPhotoTimeOffset(img_path, time_str, logger, use_tbp,
                       flickr_api, flickr_sources):
  """ Returns the number of seconds the image time runs before the time
      described in time_str. """
  photo = photohandler.PhotoList([img_path], 0, logger, use_tbp, flickr_api)
  photo.loadMetadata(True, False, False, False, flickr_sources, 0)
  photo.cleanup(check_date = True)
  if (len(photo) == 0):
    raise NoImageError("No valid image specified")
  photo_date = photo[0].getDate()
  ref_date = time.mktime(time.strptime(time_str, "%Y-%m-%d-%H-%M-%S"))
  if (photo_date):
    return int(photo_date - ref_date)
  else:
    return None

def formatUTCOffset(seconds):
  """ Returns a string with the UTC in seconds properly formatted. """
  sign = (seconds < 0) and "-" or "+"
  hrs  = abs(seconds) / 3600
  mins = (abs(seconds) % 3600) / 60
  return "UTC%s%02d:%02d" % (sign, hrs, mins)
